#!/usr/bin/env node
import fs from 'node:fs'
import { program } from 'commander'
import dotenv from 'dotenv'
import * as secondaryAccountService from '../services/secondaryAccount.js'
import * as userService from '../services/common/user.js'
import * as contextService from '../services/common/context.js'

dotenv.config({ path: '.env' })
dotenv.config({ path: '.env.defaults' })

const contextAdmin = {
  admin: {
    name: 'oxadmin',
    password: 'secret',
    display_name: 'context admin',
    sur_name: 'admin',
    given_name: 'context',
    email1: `oxadmin@${process.env.MX_DOMAIN}`,
    primaryEmail: `oxadmin@${process.env.MX_DOMAIN}`,
    login: 'oxadmin'
  }
}

function checkEnvironmentVariables () {
  const neededEnvironmentVariables = ['PROVISIONING_URL', 'E2E_ADMIN_USER', 'E2E_ADMIN_PW', 'MX_DOMAIN']
  for (const envVar of neededEnvironmentVariables) {
    if (!process.env[envVar]) {
      console.error(`Environment variable ${envVar} is required`)
      process.exit(1)
    }
  }
}

async function getOrCreateContext (data) {
  const contextData = { ...{ maxQuota: -1 }, ...data }
  let context
  try {
    const existingContext = await contextService.list(contextData.id || contextData.name)

    if (existingContext?.length > 0) {
      console.log('Context already exists, using existing one')
      context = existingContext[0]
    } else {
      console.log('Creating new context')
      context = await contextService.create(contextData)
    }
    return {
      ...context,
      ...contextAdmin
    }
  } catch (error) {
    console.error('Error in context operation:', error)
    throw error
  }
}

async function getOrCreateUser (context, userData) {
  try {
    const existingUser = await userService.list(context, userData.display_name)
    if (existingUser?.length > 0) {
      console.log('User already exists, using existing one')
      return existingUser[0]
    }
    console.log('Creating new user')
    return await userService.create(context, userData)
  } catch (error) {
    console.error('Error in user operation:', error)
    throw error
  }
}

async function getOrCreateSecondaryAccount (accountData, context, users) {
  try {
    const existingAccounts = (await secondaryAccountService.list(context, users)).filter(acc => acc.name === accountData.name)
    if (existingAccounts?.length > 0) {
      const usersWithoutAccount = users.filter(user => {
        const hasAccount = existingAccounts.some(acc => acc.userId === user.id)
        return !hasAccount
      })
      if (usersWithoutAccount.length === 0) return existingAccounts[0]
      return await secondaryAccountService.create(accountData, context, usersWithoutAccount)
    }
    console.log('Creating new secondary account')
    return await secondaryAccountService.create(accountData, context, users)
  } catch (error) {
    console.error('Error in secondary account operation:', error)
    throw error
  }
}

async function provisionFromFile (configPath) {
  checkEnvironmentVariables()
  let config
  const createdUsers = []

  try {
    config = JSON.parse(fs.readFileSync(configPath, 'utf8'))
  } catch (err) {
    console.error('Failed to read config file:', err)
    process.exit(1)
  }

  for (const contexts of config.contexts || []) {
    try {
      const contextData = contexts.data
      const context = await getOrCreateContext(contexts.data.name)

      if (contextData.capabilities) {
        try {
          await contextService.changeCapabilities(context.id, contextData.capabilities, undefined)
          console.log('Changed capabilities for context:', context.id)
        } catch (capError) {
          console.error('Error changing capabilities for context', context.id, capError)
        }
      }

      if (contextData.config) {
        try {
          await contextService.change({ id: context.id, userAttributes: contextData.config })
          console.log('Changed config for context:', context.id)
        } catch (cfgError) {
          console.error('Error changing config for context:', context.id, cfgError)
        }
      }

      for (const userData of (contexts.users || [])) {
        try {
          const user = await getOrCreateUser(context, userData)
          createdUsers.push(user)
          console.log('Provisioned user:', user && user.id)
        } catch (userError) {
          console.error('Error provisioning user', userData.display_name, userError)
        }
      }

      for (const secondaryAccount of (contexts.secondaryAccounts || [])) {
        try {
          const userIds = createdUsers.map(user => ({ id: user.id }))
          const secAccount = await getOrCreateSecondaryAccount(secondaryAccount, context, userIds)
          console.log('Provisioned secondary account:', secAccount?.primaryAddress)
        } catch (accError) {
          console.error('Error provisioning secondary account for context', context.id, accError)
        }
      }
    } catch (ctxError) {
      console.error('Error provisioning context', contexts.name, ctxError)
    }
  }
}

program
  .version('1.0.0')
  .description('OX Provisioning Tool')

program
  .option('-f, --file <path>', 'Path to provisioning JSON config file', './provisioning.json')
  .action(async (options) => {
    await provisionFromFile(options.file)
  })

program
  .command('create <resourceType>')
  .description('Create a resource (context, user, account)')
  .option('-n, --name <name>', 'Name (for context, user or account)')
  .option('-c, --context-id <id>', 'Context ID (for user or account)')
  .option('-e, --email <email>', 'Email address (for user and account)')
  .option('-q, --quota <quota>', 'Max quota in MB (for context)')
  .option('--capabilities <capabilities>', 'Comma-separated capabilities (for context)')
  .option('-p, --password <password>', 'Password (for user)', 'secret')
  .option('-g, --given-name <name>', 'Given name (for user)', 'Test')
  .option('-s, --sur-name <name>', 'Surname (for user)', 'User')
  .option('-d, --display-name <name>', 'Display name (for user)')
  .option('-u, --users <ids>', 'Comma-separated user IDs (for account)')
  .action(async (resourceType, options) => {
    checkEnvironmentVariables()

    try {
      switch (resourceType) {
        case 'context': {
          const contextData = {
            name: options.name,
            maxQuota: options.quota ? parseInt(options.quota) : -1
          }
          const context = await getOrCreateContext(contextData)

          if (options.capabilities) {
            console.log('Setting capabilities for context:', options.capabilities)
            await contextService.changeCapabilities(context.id, options.capabilities, undefined)
            console.log('Capabilities set for context:', context.id)
          }
          break
        }
        case 'user': {
          if (!options.contextId || !options.email) {
            console.error('--context-id and --email is required for user creation')
            process.exit(1)
          }

          const userData = {
            name: options.name,
            primaryEmail: options.email,
            email1: options.email,
            given_name: options.givenName || 'Test',
            sur_name: options.surName || 'User',
            display_name: options.displayName || 'Test User',
            password: options.password || 'secret'
          }

          const context = await getOrCreateContext({ id: options.contextId })
          await getOrCreateUser(context, userData)

          break
        }
        case 'account': {
          if (!options.contextId || !options.name || !options.email || !options.users) {
            console.error('--name, --email, --context-id, and --users are required for account creation')
            process.exit(1)
          }

          const accountData = {
            name: options.name,
            login: options.email,
            primaryAddress: options.email,
            mailEndpointSource: 'primary',
            transportEndpointSource: 'primary',
            personal: options.name
          }

          const contextData = await contextService.get(options.contextId)
          const context = { ...contextData, ...contextAdmin }
          const userIds = options.users.split(',').map(id => ({ id }))
          await getOrCreateSecondaryAccount(accountData, context, userIds)
          break
        }
        default:
          console.error('Unknown resource type. Must be one of: context, user, account')
          process.exit(1)
      }
    } catch (error) {
      console.error(`Error creating ${resourceType}:`, error)
      process.exit(1)
    }
  })

program
  .command('delete <resourceType>')
  .description('Delete a resource (context, user, account)')
  .option('--id <id>', 'ID of the resource (for context or user)')
  .option('-c, --context-id <id>', 'Context ID (for user or account)')
  .option('-e, --email <email>', 'Primary address (for account)')
  .option('-u, --users <ids>', 'Comma-separated user IDs (for account)')
  .action(async (resourceType, options) => {
    checkEnvironmentVariables()

    try {
      switch (resourceType) {
        case 'context':
          if (!options.id) {
            console.error('--id is required for context deletion')
            process.exit(1)
          }
          await contextService.remove(options.id)
          console.log('Context deleted successfully:', options.id)
          break

        case 'user': {
          if (!options.contextId || !options.id) {
            console.error('--context-id and --id are required for user deletion')
            process.exit(1)
          }
          const contextData = await contextService.get(options.contextId)
          const context = { ...contextData, ...contextAdmin }
          await userService.remove(context, options.id)
          console.log('User deleted successfully:', options.id)
          break
        }
        case 'account': {
          if (!options.contextId || !options.email) {
            console.error('--context-id and --primary-address are required for account deletion')
            process.exit(1)
          }
          const contextData = await contextService.get(options.contextId)
          const context = { ...contextData, ...contextAdmin }
          const users = options.users ? options.users.split(',').map(id => ({ id })) : []
          await secondaryAccountService.remove(options.email, context, users)
          console.log('Secondary account deleted successfully:', options.primaryAddress)
          break
        }
        default:
          console.error('Unknown resource type. Must be one of: context, user, account')
          process.exit(1)
      }
    } catch (error) {
      console.error(`Error deleting ${resourceType}:`, error)
      process.exit(1)
    }
  })

program.parse(process.argv)
