/**
 * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OX App Suite. If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
 *
 * Any use of the work other than as authorized under this license or copyright law is prohibited.
 */

import { createClientAsync, logSoapError } from '../../soap.js'

const OXContextService = createClientAsync('OXContextService').then(client => client)

/**
 * This function retrieves the default context from the OXContextService.
 * @returns {Promise<Object>} The first result from the listAsync method call.
 */
export async function getDefault () {
  return (await (await OXContextService).listAsync({ search_pattern: 'defaultcontext' }))[0]
}

/**
 * This function removes the context with the specified ID.
 * @param {number} id The ID of the context to remove.
 * @returns {Promise<Boolean>} Returns true if it could remove the context and false if it could not
 */
export async function remove (id) {
  return !!await (await OXContextService).deleteAsync({ ctx: { id } }).then(() => true).catch(logSoapError)
}

/**
 * This function creates a new context.
 * @param {Object} ctx The context to create.
 * @param {Object} adminUser The admin user of the context.
 *    Defaults to:
 *    ```JSON
 *    {
        name: 'oxadmin',
        password: 'secret',
        display_name: 'context admin',
        sur_name: 'admin',
        given_name: 'context',
        email1: `oxadmin@${process.env.MX_DOMAIN}`,
        primaryEmail: `oxadmin@${process.env.MX_DOMAIN}`
      }
      All properties are needed and will be inserted if not provided.
      ```
 * @returns {Promise<Object>} The created context.
 */
export async function create (ctx = {}, adminUser = {}) {
  return await (await OXContextService)
    .createAsync({
      ctx,
      admin_user: Object.assign({
        name: 'oxadmin',
        password: 'secret',
        display_name: 'context admin',
        sur_name: 'admin',
        given_name: 'context',
        email1: `oxadmin@${process.env.MX_DOMAIN}`,
        primaryEmail: `oxadmin@${process.env.MX_DOMAIN}`
      }, adminUser)
    })
    .then(async context => {
      await changeModuleAccessByName(context.id, 'all')
      return context
    })
    .catch(e => {
      logSoapError(e)
      throw e
    })
}

/**
 * This function changes the module access of the specified context.
 * @param {number} id The ID of the context.
 * @param {string} access_combination_name The name of the access combination to set.
 * @returns {Promise<void>}
 */
// eslint-disable-next-line camelcase
export async function changeModuleAccessByName (id, access_combination_name) {
  // eslint-disable-next-line camelcase
  return await (await OXContextService).changeModuleAccessByNameAsync({ ctx: { id }, access_combination_name })
}

/**
 * This function changes the capabilities of the specified context.
 * @param {number} id The ID of the context.
 * @param {string} capsToAdd The capabilities to add.
 * @param {string} capsToRemove The capabilities to remove.
 * @returns {Promise<void>}
 */
export async function changeCapabilities (id, capsToAdd, capsToRemove) {
  const data = {}
  if (capsToAdd) data.capsToAdd = capsToAdd
  if (capsToRemove) data.capsToRemove = capsToRemove
  return await (await OXContextService).changeCapabilitiesAsync({ ctx: { id }, ...data })
}

/**
 * This function retrieves the module access of the specified context.
 * @param {number} id The ID of the context.
 * @returns {Promise<Object>} The module access of the specified context.
 */
export async function getModuleAccess (id) {
  return await (await OXContextService).getModuleAccessAsync({ ctx: { id } })
}
/**
 * This function changes the module access of the specified context.
 * @param {number} id The ID of the context.
 * @param {Object} moduleAccess The module access to set.
 * @returns {Promise<void>}
 */
export async function changeModuleAccess (id, moduleAccess) {
  const currentAccess = await getModuleAccess(id)
  return await (await OXContextService).changeModuleAccessAsync({ ctx: { id }, access: { ...currentAccess, ...moduleAccess } })
}

/**
 * Search for contexts.
 * @param {String} searchPattern The pattern to search for
 * @param {Object} options Additional options
 * @param {Boolean} options.excludeDisabled Exclude disabled contexts from the search results (default: true)
 * @returns {Promise<Array<Object>>} The list of contexts that match the search pattern.
 */
export async function list (searchPattern, { excludeDisabled } = { excludeDisabled: true }) {
  return await (await OXContextService).listAsync({ search_pattern: searchPattern, exclude_disabled: excludeDisabled })
}

/**
 * This function retrieves the context with the specified ID.
 * @param {number} id The ID of the context to retrieve.
 * @returns {Promise<Object>} The context with the specified ID.
 */
export async function get (id) {
  return await (await OXContextService).getDataAsync({ ctx: { id } })
}

/**
 * This function changes the context with the specified ID.
 * @param {Object} ctx The context to change.
 * @returns {Promise<void>}
 */
export async function change (ctx) {
  return await (await OXContextService).changeAsync({ ctx })
}
