/**
 * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OX App Suite. If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
 *
 * Any use of the work other than as authorized under this license or copyright law is prohibited.
 */

import { createClientAsync, logSoapError } from '../../soap.js'

const OXUserService = createClientAsync('OXUserService').then(client => client)

/**
 * This function removes the user with the specified context and user ID.
 * @param {Object} context The context of the user to be removed.
 * @param {number} userId The ID of the user to be removed.
 * @returns {Promise<Boolean>} Returns true if it could remove the user and false if it could not
 */
export async function remove (context, userId) {
  return !!await (await OXUserService).deleteAsync({
    ctx: { id: context.id },
    user: { id: userId },
    auth: context.admin
  }).then(() => true).catch(logSoapError)
}

/**
 * This function creates a new user.
 * @param {Object} context The context of the user to be created.
 * @param {Object} usrdata The user to create.
 * @returns {Promise<Object>} The created user.
 */
export async function create (context, usrdata) {
  return await (await OXUserService).createAsync({
    ctx: { id: context.id },
    usrdata,
    auth: context.admin
  })
}

/**
 * This function changes the context with the specified ID.
 * @param {Object} context The context of the user to be changed.
 * @param {Object} usrdata The user to change.
 * @returns {Promise<void>}
 */
export async function change (context, usrdata) {
  return await (await OXUserService).changeAsync({
    ctx: { id: context.id },
    usrdata,
    auth: context.admin
  })
}

export async function changeByModuleAccess (context, userId, currentAccess, moduleAccess) {
  return await (await OXUserService).changeByModuleAccessAsync({
    ctx: { id: context.id },
    moduleAccess: Object.assign({}, currentAccess, moduleAccess),
    user: { id: userId },
    auth: context.admin
  })
}

export async function changeByModuleAccessName (context, userId, accessCombinationName) {
  return await (await OXUserService).changeByModuleAccessNameAsync({
    ctx: { id: context.id },
    access_combination_name: accessCombinationName,
    user: { id: userId },
    auth: context.admin
  }).catch(error => console.error('Module access change error:', error))
}

export async function getModuleAccess (context, userId) {
  return await (await OXUserService).getModuleAccessAsync({
    ctx: { id: context.id },
    user: { id: userId },
    auth: context.admin
  }).then(
    (res) => res
  )
}

export async function changeCapabilities (context, userId, capsToAdd, capsToRemove) {
  const data = {
    ctx: { id: context.id },
    user: { id: userId },
    auth: context.admin
  }
  if (capsToAdd) data.capsToAdd = capsToAdd
  if (capsToRemove) data.capsToRemove = capsToRemove
  return await (await OXUserService).changeCapabilitiesAsync(data)
}

export async function exists (context, userId) {
  return await (await OXUserService).existsAsync({
    ctx: { id: context.id },
    user: { id: userId },
    auth: context.admin
  })
}

export async function list (context, searchPattern, includeGuests = false) {
  return await (await OXUserService).listAsync({
    ctx: { id: context.id },
    search_pattern: searchPattern,
    include_guests: includeGuests,
    auth: context.admin
  })
}
