/**
 * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OX App Suite. If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
 *
 * Any use of the work other than as authorized under this license or copyright law is prohibited.
 */

import { performance, PerformanceObserver } from 'node:perf_hooks'
import * as SOAP from 'soap'
import dotenv from 'dotenv'
import pRetry, { AbortError as RetryAbortError } from 'p-retry'

// Set AbortError correctly
const AbortError = RetryAbortError

dotenv.config({ path: '.env' })
dotenv.config({ path: '.env.defaults' })

// This flag enables debug output including timing information.
const debug = process.env.DEBUG_SOAP === 'true'

// This URL is used to create the SOAP client.
const provisioningUrl = String(process.env.PROVISIONING_URL).replace(/\/$/, '')

// This user is used to authenticate against the provisioning API.
const provisioningAuth = {
  auth: {
    login: process.env.E2E_ADMIN_USER,
    password: process.env.E2E_ADMIN_PW
  }
}

/**
 * This code creates a PerformanceObserver instance which asynchronously observes
 * performance measurement events. It logs the name of the event (which is the name
 * of the SOAP method in this context) and the time it took to execute in milliseconds.
 *
 * This is used for performance tracking of the SOAP methods.
 * @see https://nodejs.org/api/perf_hooks.html
 * @params {PerformanceObserverEntryList} items The list of performance entries.
 * @returns {void}
 */
const obs = new PerformanceObserver((items) => {
  items.getEntries().forEach((entry) => {
    if (debug) console.log(`${entry.name} took ${Math.round(entry.duration)}ms`)
  })
})
obs.observe({ entryTypes: ['measure'] })

async function logSoapError (e) {
  console.error(e?.originalError?.root?.Envelope?.Body?.Fault?.faultstring || e.message)
}

/**
 * This function checks if the specified error should abort the retry.
 * @param {Error} error The error to check.
 * @returns {boolean} True if the error should abort the retry, false otherwise.
 **/
function shouldAbortRetry (error) {
  try {
    const fault = error.root.Envelope.Body.Fault
    const details = fault.detail
    const blockedFaultStrings = [
      /Context \d+ already exists/,
      /Authentication failed/,
      /Context \d+ does not exist/,
      /CloudPluginException: username .* already exists/,
      /CloudPluginException: context name must begin/,
      /The displayname is already used/,
      /already exists in this context/,
      /Shared Domain already exists/,
      /Shared Domain already in use/,
      /No such user/,
      /Mandatory fields in context not set/,
      /A mapping with login info .* already exists/,
      /Id must not be set if pre-assembled context should be used/,
      /Unmarshalling Error/
    ]
    const blockedExceptions = [
      'ContextExistsException',
      'InvalidCredentialsException',
      'NoSuchContextException'
    ]
    // return "details object contains any of the blocked items"
    return blockedFaultStrings.some(msg => msg.test(fault.faultstring)) ||
      blockedExceptions.some(exception => Object.hasOwnProperty.call(details, exception))
  } catch (e) {
    // some other error which is never blocked
    return false
  }
}

/**
 * This function creates a SOAP client for the specified service type.
 * @param {string} type The name of the service type.
 * @returns {Promise<Object>} The SOAP client.
 */
async function createClientAsync (type) {
  const startMark = `${type}-start`
  const endMark = `${type}-end`
  performance.mark(startMark)
  const endpoint = `${provisioningUrl}/webservices/${type}`
  const url = `${endpoint}/?wsdl`
  const client = await SOAP.createClientAsync(url, {
    endpoint,
    suppressStack: true,
    wsdl_options: {
      forever: true
    },
    gzip: true
  })

  // https://stackoverflow.com/questions/30740415/namespace-for-array-field-in-node-soap-client-node-js
  client.wsdl.definitions.xmlns.ns1 = 'http://dataobjects.soap.admin.openexchange.com/xsd'
  client.wsdl.xmlnsInEnvelope = client.wsdl._xmlnsMap()
  performance.mark(endMark)
  performance.measure(`  ⏱ SOAP: ${type} -> createClientAsync`, startMark, endMark)

  // This proxy effectively adds error handling, authentication and performance measurements to all methods of the SOAP client.
  return new Proxy(client, {
    get (target, prop, receiver) {
      const origMethod = Reflect.get(target, prop, receiver)
      if (typeof origMethod === 'function') {
        return async function (options, clientOptions, ...args) {
          const startMark = `${prop}-start`
          const endMark = `${prop}-end`
          performance.mark(startMark)
          const soapOptions = { ...provisioningAuth, ...options }
          if (soapOptions.auth) {
            // only send login and password instead of complete admin object.
            // this can fail because of ambiguous namespacing
            soapOptions.auth = { login: soapOptions.auth.login, password: soapOptions.auth.password }
          }

          const result = await pRetry(() => origMethod.apply(this, [soapOptions, { timeout: 10000, ...clientOptions }, ...args]), {
            retries: 3,
            onFailedAttempt: async error => {
              if (shouldAbortRetry(error)) throw new AbortError(error)
              console.log(`Attempt ${error.attemptNumber} failed. There are ${error.retriesLeft} retries left.`)
            }
          }).catch(e => {
            const soapError = e?.originalError?.root?.Envelope?.Body?.Fault
            throw new Error(soapError?.faultstring || e.message)
          })
          performance.mark(endMark)
          performance.measure(`  ⏱ SOAP: ${type} -> ${String(prop)}`, startMark, endMark)
          // Return only the first result from the SOAP method call, as we don't need the SOAP envelope and other stuff.
          if (!result || !result[0]) return
          return result[0]?.return
        }
      } else {
        return origMethod
      }
    }
  })
}

export { createClientAsync, logSoapError }
